////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#ifndef CRNLP_GENERAL_H 
#define CRNLP_GENERAL_H

#include <stdbool.h>
#include <stdint.h>
#include "nesting.h"
#include "cycles.h"
#include "queue.h"
#include "setup.h"

typedef struct {
    resource_mask_t       requested;
    uint32_t              length;
    uint32_t              request_id;
    uint_fast8_t volatile enabled; // set to indicate satisfaction
    uint8_t               _buf[47]; // separate nodes per cache line
} crnlp_general_request_t;

// A request Ri will correspond to |Di| nodes
typedef struct crnlp_general_node_t {
    // Time the request will be satisfied
    uint64_t start_time;

    // Doubly-LL pointers
    volatile struct crnlp_general_node_t * volatile prev;
    volatile struct crnlp_general_node_t * volatile next;

    // Request information
    crnlp_general_request_t *request;

} crnlp_general_node_t;

typedef struct {
    // Head and tail pointers
    volatile crnlp_general_node_t * volatile head;
    volatile crnlp_general_node_t * volatile tail;

} crnlp_general_queue_t;

typedef struct
{
    // Set of requests
    __attribute__((aligned(64))) crnlp_general_request_t requests[QUEUE_LENGTH];

    // Pool of available nodes, MAX_REQUESTED per request
    crnlp_general_node_t nodes[QUEUE_LENGTH][MAX_REQUESTED];

    // Set of resource queues, one per resource
    volatile crnlp_general_queue_t queues[RES_COUNT];

    // Count of total pending requests
    uint_fast8_t pending_requests;

    // Mutex lock for the overall lock data structures
    mcs_lock_t sublock;

} crnlp_general_lock_t;

// Res-queue C-RNLP initialization, lock, and unlock
void crnlp_general_init(crnlp_general_lock_t* lock);

uint64_t *crnlp_general_lock(crnlp_general_lock_t* lock, uint8_t ind);
uint64_t *crnlp_general_unlock(crnlp_general_lock_t* lock, uint8_t ind);

#endif // CRNLP_GENERAL_H
