#!/bin/bash

# Set up experiment data
source setup.sh

# Generate data
folder="results"

# Constant:
# * CS alpha (fraction not at L_max)
# * CS beta (fraction of L_max for shorter requests)
cs_alpha=0
cs_beta=100

# Varying:
# * # tasks
# * # resources
# * # resources per request
# * critical section length
# * algorithm

# Measure:
# * overhead
# * blocking

for cslength in 1 20 40 60 80 100
do
    for numresourcesnested in 1 2 4 6 8 10
    do
        for numresources in 16 32 64
        do
            for alg in 1 2 3 4 5 6 7 8 9 10 11 12
            do
                if [[ (($alg -eq 4) || ($alg -eq 9)) ]]
                then
                    # Static global lock server - need 1 extra core for the server
                    taskcounts=(${taskcounts_sg_server[@]})
                    corecounts=(${corecounts_sg_server[@]})
                    exe="./../main_xs"
                elif [[ (($alg -eq 5) || ($alg -eq 10)) ]]
                then
                    # Static local lock server(s) - need 1 extra core per socket for the server
                    taskcounts=(${taskcounts_sl_server[@]})
                    corecounts=(${corecounts_sl_server[@]})
                    exe="./../main_xs"
                else
                    # Floating or no lock server(s) - one core per task
                    taskcounts=(${taskcounts_no_server[@]})
                    corecounts=(${corecounts_no_server[@]})
                    exe="./../main"
                fi

                for (( i = 0; i < ${#taskcounts[@]}; i++ ))
                do
                    numtasks=${taskcounts[$i]}
                    numcores=${corecounts[$i]}

                    currenttime="`date +%H:%M:%S`"
                    echo "[$currenttime] n=$numtasks ($numcores cores), q=$numresources, cslength=$cslength, k=$numresourcesnested, alg=$alg"

                    $exe $numcores $numresourcesnested $numresources $cslength $cs_alpha $cs_beta $alg $folder/$alg/$numtasks-$numresources-$numresourcesnested-$cslength-$alg.csv
                done
            done
        done
    done
done
