////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "cycles.h"
#include "crnlp_general_slls.h"
#include "nesting.h"
#include "mem.h"
#include "queue.h"

void crnlp_general_slls_init(crnlp_general_slls_lock_t* lock)
{
    int i, j;

    for (i = 0; i < NUM_SERVERS; i++)
    {
        // Initialize resource queues
        for (j = 0; j < RES_COUNT; j++)
        {
            lock->servers[i].queues[j].head = NULL;
            lock->servers[i].queues[j].tail = NULL;
        }

        // Initialize the requests
        for (j = 0; j < QUEUE_LENGTH; j++)
        {
            lock->servers[i].requests[j].requested = 0;
            lock->servers[i].requests[j].length = 0;
            lock->servers[i].requests[j].request_id = i;
            lock->servers[i].requests[j].enabled = 0;
        }

        lock->servers[i].pending_requests = 0;
        lock->servers[i].satisfied_requests = 0;
        lock->servers[i].enabled = 0;

        // Initialize the RR lock logic components
        lock->servers[i].rr_wave = 0;
    }

    // Initialize the RR lock logic components
    lock->S = 0;
}

uint64_t *crnlp_general_slls_lock(crnlp_general_slls_lock_t* lock, uint8_t ind)
{
    int server_id;
    crnlp_general_slls_request_t *self;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    // Determine which server should handle this request
    server_id = ind / (QUEUE_LENGTH / NUM_SERVERS);

    self = &(lock->servers[server_id].requests[ind]);

    self->request_id = ind;
    self->enabled = 0;
    self->done = 0;

    mb();

    self->needs_service = 1;
    while (self->needs_service != 0)
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[0] += end - start;
    start = get_cycles();

    // Wait until satisfied
    while (self->enabled == 0)
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[1] += end - start;

    return ret;
}

uint64_t *crnlp_general_slls_unlock(crnlp_general_slls_lock_t* lock, uint8_t ind)
{
    int server_id;
    crnlp_general_slls_request_t *self;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    // Determine which server should handle this request
    server_id = ind / (QUEUE_LENGTH / NUM_SERVERS);

    self = &(lock->servers[server_id].requests[ind]);

    self->done = 1;

    mb();

    self->needs_service = 1;
    while (self->needs_service != 0)
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}
