////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "cycles.h"
#include "crnlp_uniform_slls.h"
#include "nesting.h"
#include "mem.h"
#include "queue.h"

void crnlp_uniform_slls_init(crnlp_uniform_slls_lock_t* lock)
{
    int i, j;

    for (i = 0; i < NUM_SERVERS; i++)
    {
        for (j=0; j < SCHEDSIZE; j++)    { lock->servers[i].enabled[j] = 0; }
        for (j=0; j < SCHEDSIZE; j++)    { lock->servers[i].blocked[j] = 0; }
        for (j=0; j < SCHEDSIZE; j++)    { lock->servers[i].ticket[j] = 0; }
        for (j=0; j < QUEUE_LENGTH; j++) { lock->servers[i].requests[j].done = 0;}
        for (j=0; j < QUEUE_LENGTH; j++) { lock->servers[i].requests[j].needs_service = 0;}

        lock->servers[i].head = 0;
        lock->servers[i].pending_requests = 0;

        memset(lock->servers[i].schedule, 0, SCHEDSIZE*sizeof(resource_mask_t));

        // Initialize the RR lock logic components
        lock->servers[i].rr_wave = 0;
    }

    // Initialize the RR lock logic components
    lock->S = 0;
}

uint64_t *crnlp_uniform_slls_lock(crnlp_uniform_slls_lock_t* lock, uint8_t ind)
{
    int server_id;
    crnlp_uniform_slls_node_t *self;
    uint_fast8_t spin_loc;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    // Determine which server should handle this request
    server_id = ind / (QUEUE_LENGTH / NUM_SERVERS);

    self = &(lock->servers[server_id].requests[ind]);

    self->done = 0;

    mb();

    self->needs_service = 1;
    while (self->needs_service != 0)
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[0] += end - start;
    start = get_cycles();

    while (lock->servers[server_id].enabled[self->spin_location] != 1)
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[1] += end - start;

    return ret;
}

uint64_t *crnlp_uniform_slls_unlock(crnlp_uniform_slls_lock_t* lock, uint8_t ind)
{
    int server_id;
    crnlp_uniform_slls_node_t *self;
    uint_fast8_t spin_loc;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    // Determine which server should handle this request
    server_id = ind / (QUEUE_LENGTH / NUM_SERVERS);

    self = &(lock->servers[server_id].requests[ind]);

    self->done = 1;

    mb();

    self->needs_service = 1;
    while (self->needs_service != 0)
    {
        cpu_relax();
    }

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

